/**************************************************************
 *                                                            *
 * description: tools for use with OpenGL                     *
 * version    : 2.5c                                          *
 * date       : 21.Feb.2003                                   *
 * modified   : 03.Mar.2005                                   *
 * author     : Jens Krger (with help form various sources)  *
 * e-mail     : jens.kruger@in.tum.de                         *
 *                                                            *
 **************************************************************/
#ifndef __OPENGLUTILS_H__
#define __OPENGLUTILS_H__

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/timeb.h>
#include <string>
#include <iostream>

#ifndef NOGLEW
	#include <GL/glew.h>
	#ifdef _WIN32
		#include <GL/wglew.h>
	#else
		#include <GL/glxew.h>
	#endif
#endif

#include <GL/glut.h>

#include "vectors.h"
#include "verbosity.h"

#define SAFE_DELETE(p)       { if(p) { delete (p);     (p)=NULL; } }
#define SAFE_DELETE_ARRAY(p) { if(p) { delete[] (p);   (p)=NULL; } }

/*
 * MACROS for timing issues. More precision than glutGet().
 */
#ifdef _WIN32
	#include <sys/timeb.h>
	#include <windows.h>
	#define TIMER_VAR(x)   DWORD ts_ ## x, te_ ## x
	#define TIMER_START(x) ts_ ## x=GetTickCount();
	#define TIMER_PROBE(x) te_ ## x=GetTickCount();
	#define TIMER_RESULT(x) (0.001*(double(te_ ## x) - double(ts_ ## x)))
#else
	#  include <unistd.h>
	#  include <sys/time.h>
	#  include <sys/times.h>
	#  include <sys/types.h>

	#  define TIMER_VAR(x)   struct timeval ts_ ## x , te_ ## x
	#  define TIMER_START(x) gettimeofday (&ts_ ## x, 0)
	#  define TIMER_PROBE(x)  gettimeofday (&te_ ## x, 0)
	#  define TIMER_RESULT(x) ((double)te_ ## x.tv_sec - (double)ts_ ## x.tv_sec + ((double)te_ ## x.tv_usec - (double)ts_ ## x.tv_usec)/1e6)

	// make sure the TCHAR marcos do not cause trouble
	# define TCHAR char
	# define _T
	# define _stprintf sprintf
	# define _tcscat strcat
#endif

// other "CONSTANT" declarations
#define PI    3.1415926535897932384626433832795 // PI 
#define root2 1.4142135623730950488016887242097 // sqrt(2)
#define root3 1.7320508075688772935274463415059	// sqrt(3)	
#define SIN(x) sin(x*0.017453292)				// sin for degrees
#define COS(x) cos(x*0.017453292)				// cos for degrees
#define ASIN(x) asin(x)*57.29577951				// arcsin for degrees
#define ACOS(x) acos(x)*57.29577951				// arccos for degrees

// OpenGL Error handling, uses define rather then a function call because of the __FILE__ and __LINE__ macros
#ifdef _DEBUG
#define checkOpenGLCreation(S,E)  { \
	if (!(E)) {\
		printf("unable to create %s \n  File: %s\n  Line: %d\n", (S), __FILE__, __LINE__);\
		char dummy; scanf( "%c",&dummy); exit(-1);\
	}\
	GLenum _errCode = glGetError();\
	if (_errCode != GL_NO_ERROR) {\
		printf("OpenGL Error found: \n  File: %s\n  Line: %d\n  (ID=%s) ERROR: %s\n",\
			     __FILE__, __LINE__, S,gluErrorString(_errCode));\
		char dummy; scanf( "%c",&dummy); exit(-1);\
	}\
}

#define checkOpenGLError(S) { \
	GLenum _errCode = glGetError();\
	if (_errCode != GL_NO_ERROR) {\
		printf("OpenGL Error found: \n  File: %s\n  Line: %d\n  (ID=%s) ERROR: %s\n",\
			     __FILE__, __LINE__, S,gluErrorString(_errCode));\
		char dummy; scanf( "%c",&dummy); exit(-1);\
	}\
}
#else
	#define checkOpenGLError(S)
	#define checkOpenGLCreation(S,E)
#endif

#define SetRG(array,index,R,G) {(array)[(index)*2] = R;(array)[(index)*2+1] = G;}
#define SetRGB(array,index,R,G,B) {(array)[(index)*3] = R;(array)[(index)*3+1] = G;(array)[(index)*3+2] = B;}
#define SetRGBA(array,index,R,G,B,A) {(array)[(index)*4] = R;(array)[(index)*4+1] = G;(array)[(index)*4+2] = B;(array)[(index)*4+3] = A;}

namespace openGLUtils {
	static double fps = 0.0;       // for fps-display. Updated every second.
	static double last;
	static double fpsCounter;
	static double elapsed_time;    // to hold queried system time

	int log2(int n);

	void fpsTimerStart();
	void fpsTimerStartFrame();
	double fpsTimerShowFPS(const char *progName);
	void fpsTimerGetFPS(double *fps);

	inline float smoothstep(float x);
	inline float scalarProd3(float a[3], float b[3]);
	inline float scalarProd4(float a[4], float b[4]);

	void MatrixMatrixProd(float a[16], float b[16]); // this function computes A=B*A   (NOT A=A*B)
	inline void MatrixVectorProd3(float M[16], double v[3]);
	inline void MatrixVectorProd3(float M[16], float v[3]);
	inline void MatrixVectorProd4(float M[16], double v[4]);
	inline void MatrixVectorProd4(float M[16], float v[4]);
	inline void printVector4(float v[4]);
	inline void printVector3(float v[3]);
	inline void printMatrix4(float m[16]);
	inline void normalize4(float v[4]);
	inline void normalize3(float v[3]);
	void glPrintString(void *font, char *string);
	inline void setVec3(float a[3], float b[3]);
	inline void swapVec3(float a[3], float b[3]);
	inline bool floatEqual(float a, float b);

	void MultEulerf(float head, float pitch, float roll);
	void GetInvertedModelview(float invM[16]); // Works only with translates & rotates !
	inline void findMinMax(int a, int b, int &min, int &max);
	inline void findMinMax(int a, int b, int c, int &min, int &max);
	inline void findMinMax(int a, int b, int c, int d, int &min, int &max);
	void saveData2BMP(char *fileName, GLubyte *image, int width, int height,bool BGR=false);
	void findNextFileName(char *prefix, char *extension, char** fileName);
	void saveWindow(char *fileName);
	void saveWindow();
	void rotate(float &rotx, float &roty, float &rotz, float rotMatrix[16], bool fluid = false, float viscosity = 1.1);
	void rotate(float &rotx, float &roty, float &rotz, FLOATMATRIX4 &rotMatrix, const bool fluid= false, const float viscosity = 1.1);

	bool isExtensionSupported(const char *extension);
	void exitWithError(char *string, int exitCode=-1);

	bool fileExits(const char* filename);
}


#endif __OPENGLUTILS_H__