/************************************************************
 *															*
 * decr     : abstract base class for all fragment based	*
 *            vector clases									*
 * version  : 1.0											*
 * author   : Jens Krger									*
 * date     : 20.05.2005									*
 * modified	: 20.05.2005									*
 * e-mail   : jens.krueger@in.tum.de						*
 *															*
 ************************************************************/

#include "clfragmentvector.h"
#include "clfragmentvector.inl" // GLSL Shaders

int	clFragmentVector::ms_iVectorCount=0;

/*******************************************************************************************
   Name:    combineLastPoints
   Purpose: the last step in the reduction process requires reduction to a specified 1x1 texture
********************************************************************************************/
void clFragmentVector::combineLastPointsGeneral(int iDepth, GLuint pMiniTex, int iType) {
	// if the vector was not "quadratic" take the aspect ratio into account for the texture coordinates
	FLOATVECTOR2 f2ReduceStep = FLOATVECTOR2((float)MAX(1,(m_memDesc.m_iHeight/m_memDesc.m_iWidth))/(iDepth*2.0f),
			                                 (float)MAX(1,(m_memDesc.m_iWidth/m_memDesc.m_iHeight))/(iDepth*2.0f));

	glViewport(0,0,1,1);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, pMiniTex, 0);

	if (iType == 0) {
		ms_pShaderCombineR->Enable();
			glUniform2f(ms_f2Scalar_CombineR,f2ReduceStep.x,f2ReduceStep.y);
			glBegin(GL_POINTS);
				glTexCoord2f(0,0);glVertex2f(0,0);
			glEnd();
		ms_pShaderCombineR->Disable();
	} else {
		ms_pShaderCombineRGBA->Enable();
			glUniform2f(ms_f2Scalar_CombineRGBA,f2ReduceStep.x,f2ReduceStep.y);
			glBegin(GL_POINTS);
				glTexCoord2f(0,0);glVertex2f(0,0);
			glEnd();
		ms_pShaderCombineRGBA->Disable();
	}
}

/*******************************************************************************************
   Name:    default destructor
   Purpose: destroyes this object
********************************************************************************************/
clFragmentVector::~clFragmentVector(void)	{
	ms_memoryMananger->releaseTexture(m_iMemID, m_pVectorTexture);
}

/*******************************************************************************************
   Name:    init
   Purpose: init for a vector of length iSize, automaticaly finds the best layout
********************************************************************************************/
void clFragmentVector::init(int iSize, bool bConstant, GLenum eFormat) {
	float fLog = openGLUtils::log2(iSize)/2.0f;
	return init(1<<(int)ceil(fLog),1<<(int)floor(fLog),bConstant,eFormat);
}

/*******************************************************************************************
   Name:    init
   Purpose: initializes this vector, should be called explitly only if default constructor
            has been used to create this object
********************************************************************************************/
void clFragmentVector::init(int iSizeX, int iSizeY, bool bConstant, GLenum eFormat) {
	m_memDesc.m_iWidth	= iSizeX;
	m_memDesc.m_iHeight	= iSizeY;
	m_memDesc.m_eFormat	= eFormat;

	// setup variables
	m_iTempActive = 0;

	// create render target, textures and surface
	m_iMemID = ms_memoryMananger->getTexture(m_memDesc, m_pVectorTexture);

	for (int i = 0;i<2;i++) {
		m_pTempTexture[i] = NULL;
		m_ID[i] = -1;
	}

	// init static variables
	InitStatic();
}

/*******************************************************************************************
   Name:    clear
   Purpose: sets all vector elements to zero
********************************************************************************************/
void clFragmentVector::clear(float fClearValue) {
	glViewport(0,0,m_memDesc.m_iWidth,m_memDesc.m_iHeight);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_pVectorTexture, 0);
		ms_pShaderClear->Enable();
			glUniform1f(ms_fScalar_Clear,fClearValue);
			RenderViewPortCover();
		ms_pShaderClear->Disable();
}

/*******************************************************************************************
   Name:    BeginScene
   Purpose: initializes the FBO for rendering i.e. 
            getting a texture from the memman and activating it
********************************************************************************************/
void clFragmentVector::BeginScene() {
	glViewport(0,0,m_memDesc.m_iWidth,m_memDesc.m_iHeight);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, getWriteTexture(), 0);

	checkOpenGLError("BeginScene");
	clMemMan::CheckFramebufferStatus();
}

/*******************************************************************************************
   Name:    EndScene
   Purpose: swaps the target texture with this texture
********************************************************************************************/
void clFragmentVector::EndScene() {
	swapSurfaces();
	checkOpenGLError("EndScene");
}

/*******************************************************************************************
   Name:    getWriteTexture
   Purpose: retrieves a target texture from the memman
********************************************************************************************/
GLuint clFragmentVector::getWriteTexture() {
	m_ID[0] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[0]);
	return m_pTempTexture[0];
}

/*******************************************************************************************
   Name:    getReadTexture
   Purpose: retrieves the current data texture of this vector
********************************************************************************************/
GLuint clFragmentVector::getReadTexture() {
	return m_pVectorTexture;
}

/*******************************************************************************************
   Name:    swapSurfaces
   Purpose: swaps the target texture with this texture
********************************************************************************************/
void clFragmentVector::swapSurfaces() {
	// swap temp and vector surface and texture
	GLuint	swapTex		= m_pVectorTexture;
	m_pVectorTexture	= m_pTempTexture[0];
	ms_memoryMananger->releaseTexture(m_ID[0],swapTex);
	m_ID[0] = -1;
}

/*******************************************************************************************
   Name:    addVector
   Purpose: adds this vector to another vector storing the result in target 
            vTarget = (this*fScal1)+(vSource*fScal2)
********************************************************************************************/
void clFragmentVector::addVector(clFragmentVector* vSource, clFragmentVector* vTarget, float fScal1, float fScal2) {
	vTarget->BeginScene();
		ms_pShaderAdd->Enable();
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);
			glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,vSource->m_pVectorTexture);
			glUniform2f(ms_f2Scalar_Add,fScal1,fScal2);

			RenderViewPortCover();

			glBindTexture(GL_TEXTURE_2D,NULL);
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,NULL);

		ms_pShaderAdd->Disable();
	vTarget->EndScene();
}

/*******************************************************************************************
   Name:    addVector
   Purpose: adds this vector to another vector storing the result in target (uses a clFloat)
            vTarget = (this*fScal)+(vSource*clfScal)
********************************************************************************************/
void clFragmentVector::addVector(clFragmentVector* vSource, clFragmentVector* vTarget, float fScal, clFloat *clfScal) {
	vTarget->BeginScene();
		ms_pShaderAddCL->Enable();
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);
			glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,vSource->m_pVectorTexture);
			glActiveTexture(GL_TEXTURE2); glBindTexture(GL_TEXTURE_2D,clfScal->m_pFloatTexture);
			glUniform1f(ms_fScalar_AddCL,fScal);

			RenderViewPortCover();

			glBindTexture(GL_TEXTURE_2D,NULL);
			glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,NULL);
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,NULL);
		ms_pShaderAddCL->Disable();
	vTarget->EndScene();
}

/*******************************************************************************************
   Name:    subtractVector
   Purpose: subtracts this vector from another vector storing the result in target (uses a clFloat)
            vTarget = (this*fScal)-(vSource*clfScal)
********************************************************************************************/
void clFragmentVector::subtractVector(clFragmentVector* vSource, clFragmentVector* vTarget, float fScal, clFloat *clfScal) {
	vTarget->BeginScene();
		ms_pShaderSubCL->Enable();
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);
			glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,vSource->m_pVectorTexture);
			glActiveTexture(GL_TEXTURE2); glBindTexture(GL_TEXTURE_2D,clfScal->m_pFloatTexture);
			glUniform1f(ms_fScalar_SubCL,fScal);

			RenderViewPortCover();

			glBindTexture(GL_TEXTURE_2D,NULL);
			glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,NULL);
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,NULL);
		ms_pShaderSubCL->Disable();
	vTarget->EndScene();
}

/*******************************************************************************************
   Name:    multiplyScalar
   Purpose: multiply this vector with a scalar (this = this*fScalar)
********************************************************************************************/
void clFragmentVector::multiplyScalar(float fScalar){
	BeginScene();
		ms_pShaderMulS->Enable();
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);
			glUniform1f(ms_fScalar_MulS,fScalar);

			RenderViewPortCover();

			glBindTexture(GL_TEXTURE_2D,NULL);		
		ms_pShaderMulS->Disable();
	EndScene();
}

/*******************************************************************************************
   Name:    copyVector
   Purpose: dublicate a vector (this.data = clvSource.data)
********************************************************************************************/
void clFragmentVector::copyVector(clFragmentVector *clvSource) {
	assert(clvSource->m_memDesc == m_memDesc);

	glViewport(0,0,m_memDesc.m_iWidth,m_memDesc.m_iHeight);

	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_pVectorTexture, 0);
		ms_pShaderCopy->Enable();
			glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,clvSource->m_pVectorTexture);
			RenderViewPortCover();
			glBindTexture(GL_TEXTURE_2D,NULL);
		ms_pShaderCopy->Disable();
}

/*******************************************************************************************
   Name:    toString
   Purpose: convert the vector to a string, outputting only the first iOutputCount values
            iOutputCount==0 (default) outputs all values
********************************************************************************************/
TCHAR* clFragmentVector::toString(int iOutputCount) {
	TCHAR* cpResult = new TCHAR[20*getSize()]; // TODO: check if 20 makes sense here
	TCHAR* cpTemp   = new TCHAR[20];
	float* pfData  = new float[getSize()];
	
	getData(pfData);

	cpResult[0] = 0;
	for (int i = 0;i<getSize();i++) {
		if (iOutputCount > 0 && iOutputCount <= i) return cpResult;

		if ((i+1)%m_memDesc.m_iWidth)
			_stprintf(cpTemp,_T("%7.5f\t "),pfData[i]);
		else
			_stprintf(cpTemp,_T("%7.5f\n"),pfData[i]);
		_tcscat(cpResult,cpTemp);
	}

	return cpResult;
}

/*******************************************************************************************
   Name:    toShortString
   Purpose: convert the vector to a string, but display only non zero entries
********************************************************************************************/
TCHAR* clFragmentVector::toShortString(int iOutputCount) {
	TCHAR* cpResult = new TCHAR[20*getSize()]; // TODO: check if 20 makes sense here
	TCHAR* cpTemp   = new TCHAR[20];
	float* pfData  = new float[getSize()];
	int iOutCounter = 0;
	
	getData(pfData);

	cpResult[0] = 0;
	for (int i = 0;i<getSize();i++) {
		if (fabs(pfData[i]) > 0.00001)	{
			_stprintf(cpTemp,_T("%i:%7.5f\t "),i, pfData[i]);
			_tcscat(cpResult,cpTemp);

			iOutCounter++;
			if (iOutputCount > 0 && iOutputCount < iOutCounter) return cpResult;
		}
	}

	return cpResult;
}
/*******************************************************************************************
   Name:    vectorOp
   Purpose: executes different vector-vector ops depending on eOpType
********************************************************************************************/
void clFragmentVector::vectorOp(CL_enum eOpType, clFragmentVector *clvSource, clFragmentVector *clvTarget, float fScal1, float fScal2) {
	switch (eOpType) {
		case CL_ADD: addVector(clvSource, clvTarget, fScal1, fScal2);
			break;
		case CL_SUB: addVector(clvSource, clvTarget, fScal1, -fScal2);
			break;
	}
}

/*******************************************************************************************
   Name:    reduceAdd
   Purpose: compute clfResult=sum(this)
********************************************************************************************/
void clFragmentVector::reduceAdd(clFloat *clfResult) {
	m_ID[0] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[0]);
	m_ID[1] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[1]);

	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);

	int iDepth = reduceAddInternal(false);
	combineLastPoints(iDepth,clfResult->m_pFloatTexture); // write directly to the clFloats texture to avoid a flipTexture call

	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,NULL);

	ms_memoryMananger->releaseTexture(m_ID[0]);
	ms_memoryMananger->releaseTexture(m_ID[1]);
}

/*******************************************************************************************
   Name:    reduceAdd
   Purpose: compute clfResult=sum(this*clvSecond)
********************************************************************************************/
void clFragmentVector::reduceAdd(clFragmentVector* clvSecond, clFloat *clfResult) {
	m_ID[0] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[0]);
	m_ID[1] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[1]);

	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);
	glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,clvSecond->m_pVectorTexture);

	int iDepth = reduceAddInternal(true);

	combineLastPoints(iDepth,clfResult->m_pFloatTexture); // write directly to the clFloats texure to avoid a flipTexture call

	glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,NULL);
	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,NULL);

	ms_memoryMananger->releaseTexture(m_ID[0]);
	ms_memoryMananger->releaseTexture(m_ID[1]);
}

/*******************************************************************************************
   Name:    reduceAdd
   Purpose: return sum(this)
********************************************************************************************/
float clFragmentVector::reduceAdd() {
	GLuint pTexture;
	int iMiniTexID = ms_memoryMananger->getTexture(clMemDescr(1,1,m_memDesc.m_eFormat,m_memDesc.m_eTarget),pTexture);
	m_ID[0] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[0]);
	m_ID[1] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[1]);

	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);

	int iDepth = reduceAddInternal(false);
	combineLastPointsGeneral(iDepth,pTexture,0);
	float result = evalLastPoint(pTexture);

	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,NULL);

	ms_memoryMananger->releaseTexture(m_ID[0]);
	ms_memoryMananger->releaseTexture(m_ID[1]);
	ms_memoryMananger->releaseTexture(iMiniTexID);
	return result;

}

/*******************************************************************************************
   Name:    reduceAdd
   Purpose: return sum(this*clvSecond)
********************************************************************************************/
float clFragmentVector::reduceAdd(clFragmentVector* clvSecond) {
	GLuint pTexture;
	int iMiniTexID = ms_memoryMananger->getTexture(clMemDescr(1,1,m_memDesc.m_eFormat,m_memDesc.m_eTarget),pTexture);
	m_ID[0] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[0]);
	m_ID[1] = ms_memoryMananger->getTexture(m_memDesc,m_pTempTexture[1]);

	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pVectorTexture);
	glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,clvSecond->m_pVectorTexture);

	int iDepth = reduceAddInternal(true);
	combineLastPointsGeneral(iDepth,pTexture,0);
	float result = evalLastPoint(pTexture);

	glActiveTexture(GL_TEXTURE1); glBindTexture(GL_TEXTURE_2D,NULL);
	glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,NULL);

	ms_memoryMananger->releaseTexture(m_ID[0]);
	ms_memoryMananger->releaseTexture(m_ID[1]);
	ms_memoryMananger->releaseTexture(iMiniTexID);
	return result;
}

/*******************************************************************************************
   Name:    reduceAddInternal
   Purpose: called by the reduceAdd methods sumFirst specifies if an other vector is to added
   in the first reduction pass
********************************************************************************************/
int clFragmentVector::reduceAddInternal(bool sumFirst) {
	int iDepth = 1;

	CProgram *currentProgram = (sumFirst) ? ms_pShaderReduceAddFirst   : ms_pShaderReduceAddRest;
	GLuint    currentScalar  = (sumFirst) ? ms_fScalar_ReduceStepFirst : ms_fScalar_ReduceStepRest;

	// a quad covering the entire viewport
	CL_TEX2D_VERTEX hQuad[4] = {{-1,-1, 0, 0},
	                            { 1,-1, 1, 0},
	                            { 1, 1, 1, 1},
	                            {-1, 1, 0, 1}};

	glViewport(0,0,m_memDesc.m_iWidth,m_memDesc.m_iHeight);

	while (iDepth*2 < m_memDesc.m_iWidth && iDepth*2 < m_memDesc.m_iHeight) {
		splitQuad(hQuad);
	
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_pTempTexture[m_iTempActive], 0);

		currentProgram->Enable();
			glUniform1f(currentScalar,1.0f/(iDepth*2.0f));
			glBegin(GL_QUADS);
				glTexCoord2f(hQuad[0].tu,hQuad[0].tv); glVertex2f(hQuad[0].x,hQuad[0].y);
				glTexCoord2f(hQuad[1].tu,hQuad[1].tv); glVertex2f(hQuad[1].x,hQuad[1].y);
				glTexCoord2f(hQuad[2].tu,hQuad[2].tv); glVertex2f(hQuad[2].x,hQuad[2].y);
				glTexCoord2f(hQuad[3].tu,hQuad[3].tv); glVertex2f(hQuad[3].x,hQuad[3].y);
			glEnd();
		currentProgram->Disable();
	
		iDepth *= 2;

		glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pTempTexture[m_iTempActive]);
		m_iTempActive = 1-m_iTempActive; // swap temp-surface

		if (iDepth == 2) {
			currentProgram = ms_pShaderReduceAddRest;
			currentScalar  = ms_fScalar_ReduceStepRest;
		}
	}
	
	// further reduce non-quadratic vectors
	if (m_memDesc.m_iWidth != m_memDesc.m_iHeight) {
		reduceInXDirection(iDepth,hQuad);	// try to reduce in x-direction
		reduceInYDirection(iDepth,hQuad);	// try to reduce in y-direction
	}

	return iDepth;
}

/*******************************************************************************************
   Name:    reduceInXDirection
   Purpose: reduces only in x-diretion, this method is required with rectangular textures
********************************************************************************************/
void clFragmentVector::reduceInXDirection(int &iDepth, CL_TEX2D_VERTEX *hQuad) {

	while (iDepth*2 < m_memDesc.m_iWidth) {
		splitQuadX(hQuad);

		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_pTempTexture[m_iTempActive], 0);

			ms_pShaderReduceAddX->Enable();
				glUniform1f(ms_fScalar_ReduceStepX,1.0f/(iDepth*2.0f));
				glBegin(GL_QUADS);
					glTexCoord2f(hQuad[0].tu,hQuad[0].tv); glVertex2f(hQuad[0].x,hQuad[0].y);
					glTexCoord2f(hQuad[1].tu,hQuad[1].tv); glVertex2f(hQuad[1].x,hQuad[1].y);
					glTexCoord2f(hQuad[2].tu,hQuad[2].tv); glVertex2f(hQuad[2].x,hQuad[2].y);
					glTexCoord2f(hQuad[3].tu,hQuad[3].tv); glVertex2f(hQuad[3].x,hQuad[3].y);
				glEnd();
			ms_pShaderReduceAddX->Disable();

		iDepth *= 2;
		glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pTempTexture[m_iTempActive]);
		m_iTempActive = 1-m_iTempActive; // swap temp-surface
	}
}

/*******************************************************************************************
   Name:    reduceInYDirection
   Purpose: reduces only in y-diretion, this method is required with rectangular textures
********************************************************************************************/
void clFragmentVector::reduceInYDirection(int &iDepth, CL_TEX2D_VERTEX *hQuad) {

	while (iDepth*2 < m_memDesc.m_iHeight) {
		splitQuadY(hQuad);

		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, m_pTempTexture[m_iTempActive], 0);

			ms_pShaderReduceAddY->Enable();
				glUniform1f(ms_fScalar_ReduceStepY,1.0f/(iDepth*2.0f));
				glBegin(GL_QUADS);
					glTexCoord2f(hQuad[0].tu,hQuad[0].tv); glVertex2f(hQuad[0].x,hQuad[0].y);
					glTexCoord2f(hQuad[1].tu,hQuad[1].tv); glVertex2f(hQuad[1].x,hQuad[1].y);
					glTexCoord2f(hQuad[2].tu,hQuad[2].tv); glVertex2f(hQuad[2].x,hQuad[2].y);
					glTexCoord2f(hQuad[3].tu,hQuad[3].tv); glVertex2f(hQuad[3].x,hQuad[3].y);
				glEnd();
			ms_pShaderReduceAddY->Disable();

		iDepth *= 2;
		glActiveTexture(GL_TEXTURE0); glBindTexture(GL_TEXTURE_2D,m_pTempTexture[m_iTempActive]);
		m_iTempActive = 1-m_iTempActive; // swap temp-surface
	}
}

/*******************************************************************************************
   Name:    splitQuadX
   Purpose: reduce a quad to half of it's size in x-dimension
********************************************************************************************/
void clFragmentVector::splitQuadX(CL_TEX2D_VERTEX *hQuad) {
	hQuad[1].x = (hQuad[1].x+1.0f)/2.0f-1.0f;
	hQuad[2].x = (hQuad[2].x+1.0f)/2.0f-1.0f;

	hQuad[1].tu /= 2.0;
	hQuad[2].tu /= 2.0;
}

/*******************************************************************************************
   Name:    splitQuadY
   Purpose: reduce a quad to half of it's size in y-dimension
********************************************************************************************/
void clFragmentVector::splitQuadY(CL_TEX2D_VERTEX *hQuad) {
	hQuad[2].y = (hQuad[2].y+1.0f)/2.0f-1.0f;
	hQuad[3].y = (hQuad[3].y+1.0f)/2.0f-1.0f;

	hQuad[2].tv /= 2.0;
	hQuad[3].tv /= 2.0;
}

/*******************************************************************************************
   Name:    splitQuad
   Purpose: reduce a quad to half of it's size in both dimensions
********************************************************************************************/
void clFragmentVector::splitQuad(CL_TEX2D_VERTEX *hQuad) {
	splitQuadX(hQuad);
	splitQuadY(hQuad);
}

/*******************************************************************************************
   Name:    InitStatic
   Purpose: loads all shaders required for all kinds of vectors
********************************************************************************************/
void clFragmentVector::InitCommonStatic() {
	ms_iVectorCount++;
	if (ms_iVectorCount > 1) return;

	ms_pShaderCopy = new CProgram(ms_vsPassThrough, ms_psCopy, TUM3D_STRING);
	ms_pShaderCopy->Enable();
		ms_pShaderCopy->SetUniformVector("g_sOtherVec", 0); // set texture
	ms_pShaderCopy->Disable();

	ms_pShaderClear = new CProgram(ms_vsPassThrough, ms_psClear, TUM3D_STRING);
	ms_pShaderClear->Enable();
		ms_fScalar_Clear=glGetUniformLocation(*ms_pShaderClear,"g_fScalar");
	ms_pShaderClear->Disable();

	ms_pShaderMulS = new CProgram(ms_vsPassThrough, ms_psMulS, TUM3D_STRING);
	ms_pShaderMulS->Enable();
		ms_fScalar_MulS=glGetUniformLocation(*ms_pShaderMulS,"g_fScalar");
		ms_pShaderMulS->SetUniformVector("g_sVec", 0); // set texture
	ms_pShaderMulS->Disable();

	ms_pShaderAdd= new CProgram(ms_vsPassThrough, ms_psAdd, TUM3D_STRING);
	ms_pShaderAdd->Enable();
		ms_f2Scalar_Add=glGetUniformLocation(*ms_pShaderAdd,"g_f2Scalar");
		ms_pShaderAdd->SetUniformVector("g_sVectorA", 0); // set texture
		ms_pShaderAdd->SetUniformVector("g_sVectorB", 1); // set texture
	ms_pShaderAdd->Disable();

	ms_pShaderAddCL= new CProgram(ms_vsPassThrough, ms_psAddCL, TUM3D_STRING);
	ms_pShaderAddCL->Enable();
		ms_fScalar_AddCL=glGetUniformLocation(*ms_pShaderAddCL,"g_fScalar");
		ms_pShaderAddCL->SetUniformVector("g_sVectorA", 0); // set texture
		ms_pShaderAddCL->SetUniformVector("g_sVectorB", 1); // set texture
		ms_pShaderAddCL->SetUniformVector("g_sScalar", 2); // set texture
	ms_pShaderAddCL->Disable();

	ms_pShaderSubCL= new CProgram(ms_vsPassThrough, ms_psSubCL, TUM3D_STRING);
	ms_pShaderSubCL->Enable();
		ms_fScalar_SubCL=glGetUniformLocation(*ms_pShaderSubCL,"g_fScalar");
		ms_pShaderSubCL->SetUniformVector("g_sVectorA", 0); // set texture
		ms_pShaderSubCL->SetUniformVector("g_sVectorB", 1); // set texture
		ms_pShaderSubCL->SetUniformVector("g_sScalar", 2); // set texture
	ms_pShaderSubCL->Disable();

	ms_pShaderReduceAddFirst= new CProgram(ms_vsReduce, ms_psReduceAddFirst, TUM3D_STRING);
	ms_pShaderReduceAddFirst->Enable();
		ms_fScalar_ReduceStepFirst=glGetUniformLocation(*ms_pShaderReduceAddFirst,"g_fReduceStep");
		ms_pShaderReduceAddFirst->SetUniformVector("g_sVectorA", 0); // set texture
		ms_pShaderReduceAddFirst->SetUniformVector("g_sVectorB", 1); // set texture
	ms_pShaderReduceAddFirst->Disable();

	ms_pShaderReduceAddRest= new CProgram(ms_vsReduce, ms_psReduceAddRest, TUM3D_STRING);
	ms_pShaderReduceAddRest->Enable();
		ms_fScalar_ReduceStepRest=glGetUniformLocation(*ms_pShaderReduceAddRest,"g_fReduceStep");
		ms_pShaderReduceAddRest->SetUniformVector("g_sVector", 0); // set texture
	ms_pShaderReduceAddRest->Disable();

	ms_pShaderReduceAddX= new CProgram(ms_vsReduce, ms_psReduceAddX, TUM3D_STRING);
	ms_pShaderReduceAddX->Enable();
		ms_fScalar_ReduceStepX=glGetUniformLocation(*ms_pShaderReduceAddX,"g_fReduceStep");
		ms_pShaderReduceAddX->SetUniformVector("g_sVector", 0); // set texture
	ms_pShaderReduceAddX->Disable();

	ms_pShaderReduceAddY= new CProgram(ms_vsReduce, ms_psReduceAddY, TUM3D_STRING);
	ms_pShaderReduceAddY->Enable();
		ms_fScalar_ReduceStepY=glGetUniformLocation(*ms_pShaderReduceAddY,"g_fReduceStep");
		ms_pShaderReduceAddY->SetUniformVector("g_sVector", 0); // set texture
	ms_pShaderReduceAddY->Disable();

	ms_pShaderCombineR= new CProgram(ms_vsCombine, ms_psCombineR, TUM3D_STRING);
	ms_pShaderCombineR->Enable();
		ms_f2Scalar_CombineR=glGetUniformLocation(*ms_pShaderCombineR,"g_f2ReduceStep");
		ms_pShaderCombineR->SetUniformVector("g_sVector", 0); // set texture
	ms_pShaderCombineR->Disable();

	ms_pShaderCombineRGBA= new CProgram(ms_vsCombine, ms_psCombineRGBA, TUM3D_STRING);
	ms_pShaderCombineRGBA->Enable();
		ms_f2Scalar_CombineRGBA=glGetUniformLocation(*ms_pShaderCombineRGBA,"g_f2ReduceStep");
		ms_pShaderCombineRGBA->SetUniformVector("g_sVector", 0); // set texture
	ms_pShaderCombineRGBA->Disable();
}

/*******************************************************************************************
   Name:    DeleteCommonStatic
   Purpose: deletes all shaders required for all kinds of vectors
********************************************************************************************/
void clFragmentVector::DeleteCommonStatic() {
	ms_iVectorCount++;
	if (ms_iVectorCount > 1) return;

	SAFE_DELETE(ms_pShaderCopy);
	SAFE_DELETE(ms_pShaderMulS);
	SAFE_DELETE(ms_pShaderAdd);
	SAFE_DELETE(ms_pShaderAddCL);
	SAFE_DELETE(ms_pShaderSubCL);
	SAFE_DELETE(ms_pShaderReduceAddFirst);
	SAFE_DELETE(ms_pShaderReduceAddRest);
	SAFE_DELETE(ms_pShaderReduceAddX);
	SAFE_DELETE(ms_pShaderReduceAddY);
	SAFE_DELETE(ms_pShaderCombineR);
	SAFE_DELETE(ms_pShaderCombineRGBA);
}